"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RabbitMQEventBus = void 0;
const amqplib_1 = require("amqplib");
const uuid_1 = require("uuid");
class MessageProcessingError extends Error {
    constructor(message, originalError) {
        super(message);
        this.originalError = originalError;
        this.name = 'MessageProcessingError';
    }
}
class RetryableError extends Error {
    constructor(message) {
        super(message);
        this.name = 'RetryableError';
    }
}
class RabbitMQEventBus {
    constructor(options) {
        this.MAX_RETRIES = 3;
        this.RETRY_DELAYS = [1000, 5000, 15000];
        this.options = {
            connection: Object.assign({ exchange: 'events', exchangeType: 'topic', deadLetterExchange: 'events.dlx' }, options.connection),
            consumer: Object.assign({ prefetch: 1, autoAck: false }, options.consumer),
            producer: Object.assign({ persistent: true, mandatory: true }, options.producer)
        };
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.connection = yield (0, amqplib_1.connect)(this.options.connection.url);
                this.channel = yield this.connection.createChannel();
                // Configuration des exchanges standards
                yield this.channel.assertExchange(this.options.connection.exchange, this.options.connection.exchangeType, { durable: true });
                // Setup de la gestion des messages morts
                yield this.setupDeadLetterHandling();
                // Configuration des event handlers
                this.connection.on('error', this.handleConnectionError.bind(this));
                this.connection.on('close', this.handleConnectionClosed.bind(this));
                this.channel.on('error', this.handleChannelError.bind(this));
                this.channel.on('close', this.handleChannelClosed.bind(this));
            }
            catch (error) {
                console.error('Failed to initialize RabbitMQ connection:', error);
                throw error;
            }
        });
    }
    setupDeadLetterHandling() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            // Configuration de l'exchange pour les messages morts
            yield this.channel.assertExchange(this.options.connection.deadLetterExchange, 'topic', { durable: true });
            // Création de la queue pour les messages morts
            const dlqName = `${this.options.connection.exchange}.dlq`;
            yield this.channel.assertQueue(dlqName, {
                durable: true,
                arguments: {
                    'x-message-ttl': 7 * 24 * 60 * 60 * 1000, // 7 jours de rétention
                    'x-max-length': 10000 // Limite de taille de la queue
                }
            });
            // Binding de la DLQ à la DLX
            yield this.channel.bindQueue(dlqName, this.options.connection.deadLetterExchange, '#' // Capturer tous les messages morts
            );
            // Mise en place du consommateur pour les messages morts
            yield this.channel.consume(dlqName, (msg) => __awaiter(this, void 0, void 0, function* () {
                var _a, _b, _c;
                if (!msg)
                    return;
                const headers = msg.properties.headers || {};
                const deathInfo = ((_a = headers['x-death']) === null || _a === void 0 ? void 0 : _a[0]) || {};
                const details = {
                    messageId: headers['x-event-id'],
                    originalExchange: deathInfo.exchange,
                    originalRoutingKey: deathInfo.routingKey,
                    failureReason: deathInfo.reason,
                    retryCount: headers['x-retry-count'] || 0,
                    content: msg.content.toString(),
                    timestamp: new Date().toISOString()
                };
                try {
                    // Traitement du message mort
                    yield this.handleDeadLetter(details);
                    // Acquittement du message
                    (_b = this.channel) === null || _b === void 0 ? void 0 : _b.ack(msg);
                }
                catch (error) {
                    console.error('Error handling dead letter:', error);
                    // En cas d'erreur, on rejette le message sans le remettre dans la queue
                    (_c = this.channel) === null || _c === void 0 ? void 0 : _c.reject(msg, false);
                }
            }), { noAck: false });
        });
    }
    handleDeadLetter(details) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            // Log détaillé du message mort
            console.error('Dead Letter Message:', Object.assign(Object.assign({}, details), { content: (_a = details.content) === null || _a === void 0 ? void 0 : _a.substring(0, 1000) // Limiter la taille du contenu dans les logs
             }));
            // Notification des équipes concernées
            yield this.notifyDeadLetter(details);
            // Possibilité d'ajouter des métriques ici
            this.incrementDeadLetterMetrics(details);
        });
    }
    notifyDeadLetter(details) {
        return __awaiter(this, void 0, void 0, function* () {
            // Implémenter ici la logique de notification
            // Exemple: Envoi d'email, webhook, alerte dans un canal Slack, etc.
            console.log('Dead Letter Notification:', {
                messageId: details.messageId,
                failureReason: details.failureReason,
                retryCount: details.retryCount,
                timestamp: details.timestamp
            });
        });
    }
    incrementDeadLetterMetrics(details) {
        // Implémenter ici la logique de métriques
        // Exemple: Incrementation de compteurs Prometheus, envoi à un service de monitoring, etc.
        console.log('Dead Letter Metrics:', {
            routingKey: details.originalRoutingKey,
            exchange: details.originalExchange,
            failureReason: details.failureReason
        });
    }
    publish(event, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            const content = Buffer.from(JSON.stringify(event));
            const timestamp = Date.now();
            try {
                yield this.channel.publish(this.options.connection.exchange, event.type, content, {
                    persistent: this.options.producer.persistent,
                    mandatory: this.options.producer.mandatory,
                    priority: options === null || options === void 0 ? void 0 : options.priority,
                    expiration: options === null || options === void 0 ? void 0 : options.expiration,
                    timestamp,
                    headers: Object.assign(Object.assign({}, options === null || options === void 0 ? void 0 : options.headers), { 'x-event-id': event.metadata.id, 'x-published-timestamp': timestamp })
                });
            }
            catch (error) {
                console.error(`Failed to publish event ${event.metadata.id}:`, error);
                throw error;
            }
        });
    }
    handleMessage(msg, handler) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a, _b;
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            const retryCount = ((_a = msg.properties.headers) === null || _a === void 0 ? void 0 : _a['x-retry-count']) || 0;
            const messageId = ((_b = msg.properties.headers) === null || _b === void 0 ? void 0 : _b['x-event-id']) || 'unknown';
            const startTime = Date.now();
            try {
                const event = JSON.parse(msg.content.toString());
                console.log(`[${messageId}] Processing message (attempt ${retryCount + 1}/${this.MAX_RETRIES + 1})`);
                yield handler(event);
                const processTime = Date.now() - startTime;
                console.log(`[${messageId}] Successfully processed message in ${processTime}ms`);
                yield this.channel.ack(msg);
            }
            catch (error) {
                const processTime = Date.now() - startTime;
                console.error(`[${messageId}] Failed to process message after ${processTime}ms:`, error);
                if (retryCount < this.MAX_RETRIES) {
                    const nextRetryDelay = this.RETRY_DELAYS[retryCount];
                    const updatedHeaders = Object.assign(Object.assign({}, msg.properties.headers), { 'x-retry-count': retryCount + 1, 'x-last-retry-timestamp': Date.now(), 'x-next-retry-delay': nextRetryDelay });
                    setTimeout(() => {
                        var _a, _b;
                        try {
                            (_a = this.channel) === null || _a === void 0 ? void 0 : _a.nack(msg, false, false);
                            console.log(`[${messageId}] Message requeued for attempt ${retryCount + 2}/${this.MAX_RETRIES + 1} after ${nextRetryDelay}ms`);
                        }
                        catch (nackError) {
                            console.error(`[${messageId}] Failed to requeue message:`, nackError);
                            (_b = this.channel) === null || _b === void 0 ? void 0 : _b.reject(msg, false);
                        }
                    }, nextRetryDelay);
                }
                else {
                    console.log(`[${messageId}] Message rejected after ${this.MAX_RETRIES} failed attempts`);
                    yield this.channel.reject(msg, false);
                }
            }
        });
    }
    subscribe(type, handler, options) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            const types = Array.isArray(type) ? type : [type];
            const queueName = (options === null || options === void 0 ? void 0 : options.queue) || (0, uuid_1.v4)();
            try {
                const { queue } = yield this.channel.assertQueue(queueName, {
                    exclusive: (_a = options === null || options === void 0 ? void 0 : options.exclusive) !== null && _a !== void 0 ? _a : !(options === null || options === void 0 ? void 0 : options.queue),
                    durable: true,
                    arguments: {
                        'x-dead-letter-exchange': this.options.connection.deadLetterExchange,
                        'x-message-ttl': 24 * 60 * 60 * 1000,
                        'x-max-length': 10000,
                        'x-overflow': 'reject-publish'
                    }
                });
                for (const eventType of types) {
                    yield this.channel.bindQueue(queue, this.options.connection.exchange, eventType);
                    console.log(`Bound queue ${queue} to event type ${eventType}`);
                }
                yield this.channel.prefetch(this.options.consumer.prefetch);
                yield this.channel.consume(queue, (msg) => __awaiter(this, void 0, void 0, function* () {
                    if (msg) {
                        yield this.handleMessage(msg, handler);
                    }
                }), {
                    consumerTag: options === null || options === void 0 ? void 0 : options.consumerTag,
                    noAck: false
                });
                console.log(`Started consuming from queue ${queue} for event types: [${types.join(', ')}]`);
            }
            catch (error) {
                console.error('Failed to setup subscription:', error);
                throw error;
            }
        });
    }
    reconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.close();
                yield this.init();
                console.log('Successfully reconnected to RabbitMQ');
            }
            catch (error) {
                console.error('Failed to reconnect:', error);
                setTimeout(() => this.reconnect(), 5000);
            }
        });
    }
    handleConnectionError(error) {
        console.error('RabbitMQ Connection Error:', error);
        this.reconnect();
    }
    handleConnectionClosed() {
        console.log('RabbitMQ Connection closed, attempting to reconnect...');
        this.reconnect();
    }
    handleChannelError(error) {
        console.error('RabbitMQ Channel Error:', error);
    }
    handleChannelClosed() {
        console.log('RabbitMQ Channel closed');
    }
    unsubscribe(consumerTag) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            try {
                yield this.channel.cancel(consumerTag);
                console.log(`Unsubscribed consumer ${consumerTag}`);
            }
            catch (error) {
                console.error(`Failed to unsubscribe consumer ${consumerTag}:`, error);
                throw error;
            }
        });
    }
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            var _a, _b;
            try {
                yield ((_a = this.channel) === null || _a === void 0 ? void 0 : _a.close());
                yield ((_b = this.connection) === null || _b === void 0 ? void 0 : _b.close());
                console.log('Closed RabbitMQ connection and channel');
            }
            catch (error) {
                console.error('Error while closing RabbitMQ connection:', error);
                throw error;
            }
        });
    }
}
exports.RabbitMQEventBus = RabbitMQEventBus;
