"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RabbitMQEventBus = void 0;
const amqplib_1 = require("amqplib");
const uuid_1 = require("uuid");
// Custom Error Classes
class MessageProcessingError extends Error {
    constructor(message, originalError) {
        super(message);
        this.originalError = originalError;
        this.name = 'MessageProcessingError';
    }
}
class RetryableError extends Error {
    constructor(message) {
        super(message);
        this.name = 'RetryableError';
    }
}
class MessageStore {
    constructor() {
        this.messages = new Map();
    }
    add(routingKey, content, options) {
        const messageId = options.messageId || (0, uuid_1.v4)();
        this.messages.set(messageId, {
            content,
            routingKey,
            options,
            timestamp: Date.now()
        });
    }
    getAll() {
        return Array.from(this.messages.values());
    }
    clear() {
        this.messages.clear();
    }
}
class RabbitMQEventBus {
    constructor(options) {
        this.isInitialized = false;
        this.reconnectAttempts = 0;
        this.messageStore = new MessageStore();
        this.boundQueues = new Set();
        // Constants
        this.MAX_RETRIES = 3;
        this.RETRY_DELAYS = [1000, 5000, 15000]; // Progressive delays in ms
        this.RECONNECT_DELAY = 5000; // 5 seconds
        this.MAX_RECONNECT_ATTEMPTS = 10;
        this.DLQ_RETENTION = 7 * 24 * 60 * 60 * 1000; // 7 days in ms
        this.QUEUE_MAX_LENGTH = 10000;
        this.options = {
            connection: Object.assign({ exchange: 'events', exchangeType: 'topic', deadLetterExchange: 'events.dlx', retryExchange: 'events.retry' }, options.connection),
            consumer: Object.assign({ prefetch: 1, autoAck: false }, options.consumer),
            producer: Object.assign({ persistent: true, mandatory: true }, options.producer),
        };
    }
    validateState() {
        if (!this.isInitialized) {
            throw new Error('EventBus not initialized. Call init() first.');
        }
        if (!this.connection || !this.channel) {
            throw new Error('Connection or channel not available.');
        }
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.connect();
                yield this.setupExchangesAndQueues();
                this.setupEventHandlers();
                this.isInitialized = true;
                console.log('RabbitMQ EventBus initialized successfully');
                // Try to republish stored messages after initialization
                yield this.republishStoredMessages();
            }
            catch (error) {
                console.error('Failed to initialize RabbitMQ EventBus:', error);
                throw error;
            }
        });
    }
    republishStoredMessages() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            const messages = this.messageStore.getAll();
            if (messages.length > 0) {
                console.log(`Attempting to republish ${messages.length} stored messages`);
                for (const message of messages) {
                    const { routingKey, content, options } = message;
                    // Only republish if we have a binding for this routing key
                    if (this.hasBindingForRoutingKey(routingKey)) {
                        try {
                            yield this.channel.publish(this.options.connection.exchange, routingKey, content, options);
                            console.log(`Successfully republished message to ${routingKey}`);
                        }
                        catch (error) {
                            console.error(`Failed to republish message to ${routingKey}:`, error);
                        }
                    }
                    else {
                        console.log(`No binding found for routing key ${routingKey}, keeping message in store`);
                    }
                }
            }
        });
    }
    hasBindingForRoutingKey(routingKey) {
        // Check if we have an exact match or a matching pattern
        return Array.from(this.boundQueues).some(binding => {
            // Convert binding pattern to regex
            const pattern = binding
                .replace('.', '\\.')
                .replace('*', '[^.]+')
                .replace('#', '.*');
            const regex = new RegExp(`^${pattern}$`);
            return regex.test(routingKey);
        });
    }
    handleUnroutableMessage(msg) {
        console.log('Message returned as unroutable, storing for later delivery:', {
            exchange: msg.fields.exchange,
            routingKey: msg.fields.routingKey,
            messageId: msg.properties.messageId
        });
        this.messageStore.add(msg.fields.routingKey, msg.content, msg.properties);
    }
    connect() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.connection = yield (0, amqplib_1.connect)(this.options.connection.url);
                this.channel = yield this.connection.createChannel();
                this.reconnectAttempts = 0;
                console.log('Connected to RabbitMQ successfully');
            }
            catch (error) {
                this.reconnectAttempts++;
                if (this.reconnectAttempts < this.MAX_RECONNECT_ATTEMPTS) {
                    console.warn(`Connection attempt ${this.reconnectAttempts} failed, retrying in ${this.RECONNECT_DELAY}ms`);
                    yield new Promise(resolve => setTimeout(resolve, this.RECONNECT_DELAY));
                    return this.connect();
                }
                throw error;
            }
        });
    }
    setupExchangesAndQueues() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                throw new Error('Channel not initialized');
            // Setup main exchange
            yield this.channel.assertExchange(this.options.connection.exchange, this.options.connection.exchangeType, { durable: true });
            // Setup retry exchange
            yield this.channel.assertExchange(this.options.connection.retryExchange, 'direct', { durable: true });
            // Setup dead letter exchange
            yield this.channel.assertExchange(this.options.connection.deadLetterExchange, 'topic', { durable: true });
            // Setup dead letter queue with safer configuration
            const dlqName = `${this.options.connection.exchange}.dlq`;
            try {
                yield this.channel.checkQueue(dlqName);
                console.log(`DLQ ${dlqName} already exists, skipping queue creation`);
            }
            catch (error) {
                if (error.code === 404) {
                    yield this.channel.assertQueue(dlqName, {
                        durable: true,
                        arguments: {
                            'x-message-ttl': this.DLQ_RETENTION,
                            'x-max-length': this.QUEUE_MAX_LENGTH,
                            'x-overflow': 'reject-publish'
                        }
                    });
                    console.log(`Created DLQ ${dlqName}`);
                }
                else {
                    throw error;
                }
            }
            yield this.channel.bindQueue(dlqName, this.options.connection.deadLetterExchange, '#');
            yield this.setupDeadLetterConsumer(dlqName);
        });
    }
    setupEventHandlers() {
        if (!this.connection || !this.channel)
            return;
        // Connection event handlers
        this.connection.on('error', this.handleConnectionError.bind(this));
        this.connection.on('close', this.handleConnectionClosed.bind(this));
        // Channel event handlers
        this.channel.on('error', this.handleChannelError.bind(this));
        this.channel.on('close', this.handleChannelClosed.bind(this));
        this.channel.on('return', this.handleUnroutableMessage.bind(this));
    }
    handleConnectionError(error) {
        return __awaiter(this, void 0, void 0, function* () {
            console.error('Connection error:', error);
            yield this.reconnect();
        });
    }
    handleConnectionClosed() {
        return __awaiter(this, void 0, void 0, function* () {
            console.warn('Connection closed, attempting to reconnect...');
            yield this.reconnect();
        });
    }
    handleChannelError(error) {
        return __awaiter(this, void 0, void 0, function* () {
            console.error('Channel error:', error);
            yield this.recreateChannel();
        });
    }
    handleChannelClosed() {
        return __awaiter(this, void 0, void 0, function* () {
            console.warn('Channel closed, attempting to recreate...');
            yield this.recreateChannel();
        });
    }
    reconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            this.isInitialized = false;
            try {
                yield this.init();
            }
            catch (error) {
                console.error('Failed to reconnect:', error);
                setTimeout(() => this.reconnect(), this.RECONNECT_DELAY);
            }
        });
    }
    recreateChannel() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                if (!this.connection)
                    throw new Error('No connection available');
                this.channel = yield this.connection.createChannel();
                yield this.setupExchangesAndQueues();
                console.log('Channel recreated successfully');
            }
            catch (error) {
                console.error('Failed to recreate channel:', error);
                yield this.reconnect();
            }
        });
    }
    setupDeadLetterConsumer(dlqName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            yield this.channel.consume(dlqName, (msg) => __awaiter(this, void 0, void 0, function* () {
                var _a, _b;
                if (!msg)
                    return;
                try {
                    const details = this.extractDeadLetterDetails(msg);
                    yield this.handleDeadLetter(details);
                    yield ((_a = this.channel) === null || _a === void 0 ? void 0 : _a.ack(msg));
                }
                catch (error) {
                    console.error('Error processing dead letter:', error);
                    yield ((_b = this.channel) === null || _b === void 0 ? void 0 : _b.reject(msg, false));
                }
            }), { noAck: false });
        });
    }
    extractDeadLetterDetails(msg) {
        var _a;
        const headers = msg.properties.headers || {};
        const deathInfo = ((_a = headers['x-death']) === null || _a === void 0 ? void 0 : _a[0]) || {};
        return {
            messageId: msg.properties.messageId || 'unknown',
            failureReason: headers['x-failure-reason'] || 'Unknown failure',
            retryCount: headers['x-retry-count'] || 0,
            timestamp: new Date().toISOString(),
            originalExchange: deathInfo.exchange || '',
            originalRoutingKey: deathInfo.routingKey || '',
            content: msg.content.toString(),
            headers: headers,
            error: headers['x-error-message'] ? new Error(headers['x-error-message']) : undefined
        };
    }
    handleDeadLetter(details) {
        return __awaiter(this, void 0, void 0, function* () {
            console.error('Dead Letter Processing:', {
                messageId: details.messageId,
                failureReason: details.failureReason,
                retryCount: details.retryCount,
                timestamp: details.timestamp
            });
            if (this.shouldAttemptRecovery(details)) {
                yield this.attemptMessageRecovery(details);
            }
            else {
                yield this.notifyDeadLetter(details);
                this.incrementDeadLetterMetrics(details);
            }
        });
    }
    shouldAttemptRecovery(details) {
        var _a;
        const recoverableErrors = ['ConnectionError', 'TimeoutError'];
        return (recoverableErrors.includes(((_a = details.error) === null || _a === void 0 ? void 0 : _a.name) || '') &&
            details.retryCount < this.MAX_RETRIES);
    }
    attemptMessageRecovery(details) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            if (!this.channel)
                throw new Error('Channel not initialized');
            try {
                yield this.channel.publish(this.options.connection.retryExchange, details.originalRoutingKey, Buffer.from(details.content), {
                    headers: Object.assign(Object.assign({}, details.headers), { 'x-retry-count': details.retryCount + 1, 'x-recovery-attempt': true, 'x-original-error': (_a = details.error) === null || _a === void 0 ? void 0 : _a.message })
                });
                console.log(`Recovery attempt initiated for message ${details.messageId}`);
            }
            catch (error) {
                console.error('Recovery attempt failed:', error);
                throw error;
            }
        });
    }
    notifyDeadLetter(details) {
        return __awaiter(this, void 0, void 0, function* () {
            console.log('Dead Letter Notification:', {
                messageId: details.messageId,
                failureReason: details.failureReason,
                retryCount: details.retryCount,
                timestamp: details.timestamp
            });
        });
    }
    incrementDeadLetterMetrics(details) {
        console.log('Dead Letter Metrics:', {
            routingKey: details.originalRoutingKey,
            exchange: details.originalExchange,
            failureReason: details.failureReason,
            retryCount: details.retryCount
        });
    }
    publish(event, options) {
        return __awaiter(this, void 0, void 0, function* () {
            this.validateState();
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            const content = Buffer.from(JSON.stringify(event));
            const messageId = event.metadata.id || (0, uuid_1.v4)();
            const timestamp = Date.now();
            try {
                const published = yield this.channel.publish(this.options.connection.exchange, event.type, content, Object.assign({ persistent: this.options.producer.persistent, mandatory: this.options.producer.mandatory, messageId,
                    timestamp, headers: Object.assign({ 'x-event-id': messageId, 'x-published-timestamp': timestamp, 'x-correlation-id': event.metadata.correlationId, 'x-source': event.metadata.source }, options === null || options === void 0 ? void 0 : options.headers) }, options));
                if (!published) {
                    throw new Error('Message was not confirmed by the broker');
                }
            }
            catch (error) {
                const wrappedError = new MessageProcessingError(`Failed to publish event ${messageId}`, error);
                console.error('Publication error:', wrappedError);
                throw wrappedError;
            }
        });
    }
    // private handleUnroutableMessage(msg: ConsumeMessage): void {
    //   console.error('Message returned as unroutable:', {
    //     exchange: msg.fields.exchange,
    //     routingKey: msg.fields.routingKey,
    //     messageId: msg.properties.messageId
    //   });
    // }
    subscribe(type, handler, options) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            this.validateState();
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            const types = Array.isArray(type) ? type : [type];
            const queueName = (options === null || options === void 0 ? void 0 : options.queue) || (0, uuid_1.v4)();
            try {
                const { queue } = yield this.channel.assertQueue(queueName, {
                    exclusive: (_a = options === null || options === void 0 ? void 0 : options.exclusive) !== null && _a !== void 0 ? _a : !(options === null || options === void 0 ? void 0 : options.queue),
                    durable: true,
                    arguments: {
                        'x-dead-letter-exchange': this.options.connection.deadLetterExchange,
                        'x-message-ttl': 24 * 60 * 60 * 1000,
                        'x-max-length': this.QUEUE_MAX_LENGTH,
                        'x-overflow': 'reject-publish'
                    }
                });
                // Bind queue to all event types and store bindings
                for (const eventType of types) {
                    yield this.channel.bindQueue(queue, this.options.connection.exchange, eventType);
                    this.boundQueues.add(eventType);
                    console.log(`Bound queue ${queue} to event type ${eventType}`);
                }
                yield this.channel.prefetch(this.options.consumer.prefetch);
                yield this.channel.consume(queue, (msg) => __awaiter(this, void 0, void 0, function* () {
                    if (!msg)
                        return;
                    yield this.handleMessage(msg, handler);
                }), { noAck: false });
                // After successful subscription, try to republish stored messages
                yield this.republishStoredMessages();
            }
            catch (error) {
                console.error(`Failed to subscribe to ${types.join(', ')}:`, error);
                throw error;
            }
        });
    }
    handleMessage(msg, handler) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            if (!this.channel)
                throw new Error('Channel not initialized');
            const messageId = msg.properties.messageId || 'unknown';
            const retryCount = ((_a = msg.properties.headers) === null || _a === void 0 ? void 0 : _a['x-retry-count']) || 0;
            const startTime = Date.now();
            try {
                const event = JSON.parse(msg.content.toString());
                console.log(`[${messageId}] Processing message (attempt ${retryCount + 1}/${this.MAX_RETRIES + 1})`);
                yield handler(event);
                const processTime = Date.now() - startTime;
                console.log(`[${messageId}] Successfully processed message in ${processTime}ms`);
                if (!this.options.consumer.autoAck) {
                    yield this.channel.ack(msg);
                }
            }
            catch (error) {
                const processTime = Date.now() - startTime;
                if (error instanceof RetryableError && retryCount < this.MAX_RETRIES) {
                    yield this.handleRetryableError(msg, error, retryCount);
                }
                else {
                    yield this.handleNonRetryableError(msg, error);
                }
                console.error(`[${messageId}] Message processing failed after ${processTime}ms:`, error);
            }
        });
    }
    handleRetryableError(msg, error, retryCount) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            const nextRetryDelay = this.RETRY_DELAYS[retryCount];
            const headers = Object.assign(Object.assign({}, msg.properties.headers), { 'x-retry-count': retryCount + 1, 'x-last-retry-timestamp': Date.now(), 'x-next-retry-delay': nextRetryDelay, 'x-last-error': error.message, 'x-error-stack': error.stack });
            try {
                yield new Promise(resolve => setTimeout(resolve, nextRetryDelay));
                const published = yield this.channel.publish(this.options.connection.retryExchange, msg.fields.routingKey, msg.content, Object.assign(Object.assign({}, msg.properties), { headers }));
                if (published) {
                    yield this.channel.ack(msg);
                    console.log(`Message requeued for retry ${retryCount + 1}/${this.MAX_RETRIES}`);
                }
                else {
                    throw new Error('Failed to publish retry message');
                }
            }
            catch (retryError) {
                console.error('Failed to handle retryable error:', retryError);
                yield this.handleNonRetryableError(msg, error);
            }
        });
    }
    handleNonRetryableError(msg, error) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            const headers = Object.assign(Object.assign({}, msg.properties.headers), { 'x-failure-reason': error.message, 'x-failed-timestamp': Date.now(), 'x-error-type': error.name, 'x-error-stack': error.stack, 'x-original-exchange': msg.fields.exchange, 'x-original-routing-key': msg.fields.routingKey });
            try {
                const published = yield this.channel.publish(this.options.connection.deadLetterExchange, 'dead.letter', msg.content, Object.assign(Object.assign({}, msg.properties), { headers }));
                if (published) {
                    yield this.channel.ack(msg);
                    console.log(`Message moved to DLQ: ${msg.properties.messageId}`);
                }
                else {
                    throw new Error('Failed to publish to DLQ');
                }
            }
            catch (dlqError) {
                console.error('Failed to move message to DLQ:', dlqError);
                yield this.channel.reject(msg, false);
            }
        });
    }
    unsubscribe(consumerTag) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel) {
                throw new Error('Channel not initialized');
            }
            try {
                yield this.channel.cancel(consumerTag);
                console.log(`Unsubscribed consumer ${consumerTag}`);
            }
            catch (error) {
                console.error(`Failed to unsubscribe consumer ${consumerTag}:`, error);
                throw error;
            }
        });
    }
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                if (this.channel) {
                    yield this.channel.close();
                    this.channel = undefined;
                }
                if (this.connection) {
                    yield this.connection.close();
                    this.connection = undefined;
                }
                this.isInitialized = false;
                console.log('RabbitMQ EventBus closed successfully');
            }
            catch (error) {
                const wrappedError = new MessageProcessingError('Failed to close RabbitMQ connection', error);
                console.error('Close error:', wrappedError);
                throw wrappedError;
            }
        });
    }
}
exports.RabbitMQEventBus = RabbitMQEventBus;
