"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeadLetterHandler = void 0;
class DeadLetterHandler {
    constructor(channel, maxRetries, retryExchange) {
        this.channel = channel;
        this.maxRetries = maxRetries;
        this.retryExchange = retryExchange;
    }
    setupDeadLetterConsumer(dlqName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            yield this.channel.consume(dlqName, (msg) => __awaiter(this, void 0, void 0, function* () {
                var _a, _b;
                if (!msg)
                    return;
                try {
                    const details = this.extractDeadLetterDetails(msg);
                    yield this.handleDeadLetter(details);
                    yield ((_a = this.channel) === null || _a === void 0 ? void 0 : _a.ack(msg));
                }
                catch (error) {
                    console.error("Error processing dead letter:", error);
                    yield ((_b = this.channel) === null || _b === void 0 ? void 0 : _b.reject(msg, false));
                }
            }), { noAck: false });
        });
    }
    extractDeadLetterDetails(msg) {
        var _a;
        const headers = msg.properties.headers || {};
        const deathInfo = ((_a = headers["x-death"]) === null || _a === void 0 ? void 0 : _a[0]) || {};
        return {
            messageId: msg.properties.messageId || "unknown",
            failureReason: headers["x-failure-reason"] || "Unknown failure",
            retryCount: headers["x-retry-count"] || 0,
            timestamp: new Date().toISOString(),
            originalExchange: deathInfo.exchange || "",
            originalRoutingKey: deathInfo.routingKey || "",
            content: msg.content.toString(),
            headers: headers,
            error: headers["x-error-message"]
                ? new Error(headers["x-error-message"])
                : undefined,
        };
    }
    handleDeadLetter(details) {
        return __awaiter(this, void 0, void 0, function* () {
            console.error("Dead Letter Processing:", {
                messageId: details.messageId,
                failureReason: details.failureReason,
                retryCount: details.retryCount,
                timestamp: details.timestamp,
            });
            if (this.shouldAttemptRecovery(details)) {
                yield this.attemptMessageRecovery(details);
            }
            else {
                yield this.notifyDeadLetter(details);
                this.incrementDeadLetterMetrics(details);
            }
        });
    }
    shouldAttemptRecovery(details) {
        var _a;
        const recoverableErrors = ["ConnectionError", "TimeoutError"];
        return (recoverableErrors.includes(((_a = details.error) === null || _a === void 0 ? void 0 : _a.name) || "") &&
            details.retryCount < this.maxRetries);
    }
    attemptMessageRecovery(details) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            if (!this.channel)
                throw new Error("Channel not initialized");
            try {
                yield this.channel.publish(this.retryExchange, details.originalRoutingKey, Buffer.from(details.content), {
                    headers: Object.assign(Object.assign({}, details.headers), { "x-retry-count": details.retryCount + 1, "x-recovery-attempt": true, "x-original-error": (_a = details.error) === null || _a === void 0 ? void 0 : _a.message }),
                });
                console.log(`Recovery attempt initiated for message ${details.messageId}`);
            }
            catch (error) {
                console.error("Recovery attempt failed:", error);
                throw error;
            }
        });
    }
    notifyDeadLetter(details) {
        return __awaiter(this, void 0, void 0, function* () {
            console.log("Dead Letter Notification:", {
                messageId: details.messageId,
                failureReason: details.failureReason,
                retryCount: details.retryCount,
                timestamp: details.timestamp,
            });
        });
    }
    incrementDeadLetterMetrics(details) {
        console.log("Dead Letter Metrics:", {
            routingKey: details.originalRoutingKey,
            exchange: details.originalExchange,
            failureReason: details.failureReason,
            retryCount: details.retryCount,
        });
    }
}
exports.DeadLetterHandler = DeadLetterHandler;
