"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MessageHandler = void 0;
const errors_1 = require("../errors");
class MessageHandler {
    constructor(channel, idempotencyStore, options, retryConfig) {
        this.channel = channel;
        this.idempotencyStore = idempotencyStore;
        this.options = options;
        this.retryConfig = retryConfig;
    }
    parseEvent(msg) {
        try {
            return JSON.parse(msg.content.toString());
        }
        catch (error) {
            throw new Error(`Failed to parse message content: ${error}`);
        }
    }
    handleMessage(msg, handler) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a, _b;
            if (!this.channel)
                throw new Error("Channel not initialized");
            const messageId = msg.properties.messageId ||
                ((_a = msg.properties.headers) === null || _a === void 0 ? void 0 : _a["x-event-id"]) ||
                "unknown";
            const retryCount = ((_b = msg.properties.headers) === null || _b === void 0 ? void 0 : _b["x-retry-count"]) || 0;
            const startTime = Date.now();
            try {
                const { processed, result } = yield this.idempotencyStore.hasBeenProcessed(messageId);
                if (processed) {
                    console.log(`[${messageId}] Message already processed, skipping`);
                    yield this.channel.ack(msg);
                    return;
                }
                const event = this.parseEvent(msg);
                console.log(`[${messageId}] Processing message (attempt ${retryCount + 1}/${this.retryConfig.maxRetries + 1})`);
                yield handler(event);
                yield this.idempotencyStore.markAsProcessed(messageId);
                const processTime = Date.now() - startTime;
                console.log(`[${messageId}] Successfully processed message in ${processTime}ms`);
                if (!this.options.consumer.autoAck) {
                    yield this.channel.ack(msg);
                }
            }
            catch (error) {
                const processTime = Date.now() - startTime;
                if (error instanceof errors_1.RetryableError && retryCount < this.retryConfig.maxRetries) {
                    yield this.handleRetryableError(msg, error, retryCount);
                }
                else {
                    yield this.handleNonRetryableError(msg, error);
                }
                console.error(`[${messageId}] Message processing failed after ${processTime}ms:`, error);
            }
        });
    }
    handleRetryableError(msg, error, retryCount) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            const nextRetryDelay = this.retryConfig.delays[retryCount];
            const headers = Object.assign(Object.assign({}, msg.properties.headers), { "x-retry-count": retryCount + 1, "x-last-retry-timestamp": Date.now(), "x-next-retry-delay": nextRetryDelay, "x-last-error": error.message, "x-error-stack": error.stack });
            try {
                yield new Promise((resolve) => setTimeout(resolve, nextRetryDelay));
                const published = yield this.channel.publish(this.retryConfig.retryExchange, msg.fields.routingKey, msg.content, Object.assign(Object.assign({}, msg.properties), { headers }));
                if (published) {
                    yield this.channel.ack(msg);
                    console.log(`Message requeued for retry ${retryCount + 1}/${this.retryConfig.maxRetries}`);
                }
                else {
                    throw new Error("Failed to publish retry message");
                }
            }
            catch (retryError) {
                console.error("Failed to handle retryable error:", retryError);
                yield this.handleNonRetryableError(msg, error);
            }
        });
    }
    handleNonRetryableError(msg, error) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            const headers = Object.assign(Object.assign({}, msg.properties.headers), { "x-failure-reason": error.message, "x-failed-timestamp": Date.now(), "x-error-type": error.name, "x-error-stack": error.stack, "x-original-exchange": msg.fields.exchange, "x-original-routing-key": msg.fields.routingKey });
            try {
                const published = yield this.channel.publish(this.retryConfig.deadLetterExchange, "dead.letter", msg.content, Object.assign(Object.assign({}, msg.properties), { headers }));
                if (published) {
                    yield this.channel.ack(msg);
                    console.log(`Message moved to DLQ: ${msg.properties.messageId}`);
                }
                else {
                    throw new Error("Failed to publish to DLQ");
                }
            }
            catch (dlqError) {
                console.error("Failed to move message to DLQ:", dlqError);
                yield this.channel.reject(msg, false);
            }
        });
    }
}
exports.MessageHandler = MessageHandler;
