"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RabbitMQEventBus = void 0;
const amqplib_1 = require("amqplib");
const uuid_1 = require("uuid");
const message_store_1 = require("./stores/message-store");
const indempotency_1 = require("./stores/indempotency");
const connection_handler_1 = require("./handlers/connection-handler");
const message_handler_1 = require("./handlers/message-handler");
const dl_handler_1 = require("./handlers/dl-handler");
const errors_1 = require("./errors");
const bus_options_validator_1 = require("../utils/bus-options-validator");
class RabbitMQEventBus {
    constructor(options) {
        this.isInitialized = false;
        this.disableReconnection = false;
        this.reconnectAttempts = 0;
        this.MAX_RETRIES = 3;
        this.RETRY_DELAYS = [1000, 5000, 15000];
        this.RECONNECT_DELAY = 5000;
        this.MAX_RECONNECT_ATTEMPTS = 10;
        this.DLQ_RETENTION = 7 * 24 * 60 * 60 * 1000;
        this.QUEUE_MAX_LENGTH = 10000;
        this.options = (0, bus_options_validator_1.validateAndMergeOptions)(options);
        this.messageStore = new message_store_1.MessageStore();
        this.boundQueues = new Set();
        this.idempotencyStore = new indempotency_1.IdempotencyStore();
        this.connectionHandler = new connection_handler_1.ConnectionHandler(this.reconnect.bind(this), this.recreateChannel.bind(this));
    }
    setDisableReconnection(value) {
        this.disableReconnection = value;
    }
    validateState() {
        if (!this.isInitialized) {
            throw new Error("EventBus not initialized. Call init() first.");
        }
        if (!this.connection || !this.channel) {
            throw new Error("Connection or channel not available.");
        }
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.connect();
                yield this.setupExchangesAndQueues();
                yield this.initializeHandlers();
                this.setupEventHandlers();
                this.isInitialized = true;
                console.log("RabbitMQ EventBus initialized successfully");
                yield this.republishStoredMessages();
            }
            catch (error) {
                console.error("Failed to initialize RabbitMQ EventBus:", error);
                throw error;
            }
        });
    }
    connect() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.connection = yield (0, amqplib_1.connect)(this.options.connection.url);
                this.channel = yield this.connection.createChannel();
                this.reconnectAttempts = 0;
                console.log("Connected to RabbitMQ successfully");
            }
            catch (error) {
                this.reconnectAttempts++;
                if (!this.disableReconnection && this.reconnectAttempts < this.MAX_RECONNECT_ATTEMPTS) {
                    console.warn(`Connection attempt ${this.reconnectAttempts} failed, retrying in ${this.RECONNECT_DELAY}ms`);
                    yield new Promise((resolve) => setTimeout(resolve, this.RECONNECT_DELAY));
                    return this.connect();
                }
                throw error;
            }
        });
    }
    initializeHandlers() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                throw new Error("Channel not initialized");
            this.messageHandler = new message_handler_1.MessageHandler(this.channel, this.idempotencyStore, this.options, {
                maxRetries: this.MAX_RETRIES,
                delays: this.RETRY_DELAYS,
                retryExchange: this.options.connection.retryExchange,
                deadLetterExchange: this.options.connection.deadLetterExchange,
            });
            this.deadLetterHandler = new dl_handler_1.DeadLetterHandler(this.channel, this.MAX_RETRIES, this.options.connection.retryExchange);
        });
    }
    setupExchangesAndQueues() {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            if (!this.channel)
                throw new Error("Channel not initialized");
            yield Promise.all([
                this.channel.assertExchange(this.options.connection.exchange, this.options.connection.exchangeType, { durable: true }),
                this.channel.assertExchange(this.options.connection.retryExchange, "direct", { durable: true }),
                this.channel.assertExchange(this.options.connection.deadLetterExchange, "topic", { durable: true })
            ]);
            const dlqName = `${this.options.connection.exchange}.dlq`;
            yield this.channel.assertQueue(dlqName, {
                durable: true,
                arguments: {
                    "x-message-ttl": this.DLQ_RETENTION,
                    "x-max-length": this.QUEUE_MAX_LENGTH,
                    "x-overflow": "reject-publish",
                },
            });
            yield this.channel.bindQueue(dlqName, this.options.connection.deadLetterExchange, "#");
            yield ((_a = this.deadLetterHandler) === null || _a === void 0 ? void 0 : _a.setupDeadLetterConsumer(dlqName));
        });
    }
    setupEventHandlers() {
        if (!this.connection || !this.channel)
            return;
        this.connection.on("error", this.connectionHandler.handleConnectionError.bind(this.connectionHandler));
        this.connection.on("close", this.connectionHandler.handleConnectionClosed.bind(this.connectionHandler));
        this.channel.on("error", this.connectionHandler.handleChannelError.bind(this.connectionHandler));
        this.channel.on("close", this.connectionHandler.handleChannelClosed.bind(this.connectionHandler));
        this.channel.on("return", this.handleUnroutableMessage.bind(this));
    }
    handleUnroutableMessage(msg) {
        console.log("Message returned as unroutable, storing for later delivery:", {
            exchange: msg.fields.exchange,
            routingKey: msg.fields.routingKey,
            messageId: msg.properties.messageId,
        });
        this.messageStore.add(msg.fields.routingKey, msg.content, msg.properties);
    }
    republishStoredMessages() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel)
                return;
            const messages = this.messageStore.getAll();
            if (messages.length === 0)
                return;
            console.log(`Attempting to republish ${messages.length} stored messages`);
            for (const message of messages) {
                const { routingKey, content, options } = message;
                if (this.hasBindingForRoutingKey(routingKey)) {
                    try {
                        yield this.channel.publish(this.options.connection.exchange, routingKey, content, options);
                        console.log(`Successfully republished message to ${routingKey}`);
                    }
                    catch (error) {
                        console.error(`Failed to republish message to ${routingKey}:`, error);
                    }
                }
            }
        });
    }
    hasBindingForRoutingKey(routingKey) {
        return Array.from(this.boundQueues).some((binding) => {
            const pattern = binding
                .replace(".", "\\.")
                .replace("*", "[^.]+")
                .replace("#", ".*");
            return new RegExp(`^${pattern}$`).test(routingKey);
        });
    }
    publish(event, options) {
        return __awaiter(this, void 0, void 0, function* () {
            this.validateState();
            if (!this.channel)
                throw new Error("Channel not initialized");
            const content = Buffer.from(JSON.stringify(event));
            const messageId = event.metadata.id || (0, uuid_1.v4)();
            const timestamp = Date.now();
            try {
                const published = yield this.channel.publish(this.options.connection.exchange, event.type, content, Object.assign({ persistent: this.options.producer.persistent, mandatory: this.options.producer.mandatory, messageId,
                    timestamp, headers: Object.assign({ 'x-event-id': messageId, 'x-published-timestamp': timestamp, 'x-correlation-id': event.metadata.correlationId, 'x-source': event.metadata.source, 'x-deduplication-id': `${event.type}-${messageId}` }, options === null || options === void 0 ? void 0 : options.headers) }, options));
                if (!published) {
                    throw new Error('Message was not confirmed by the broker');
                }
            }
            catch (error) {
                throw new errors_1.MessageProcessingError(`Failed to publish event ${messageId}`, error);
            }
        });
    }
    subscribe(type, handler, options) {
        return __awaiter(this, void 0, void 0, function* () {
            var _a;
            this.validateState();
            if (!this.channel || !this.messageHandler) {
                throw new Error("Channel or message handler not initialized");
            }
            const types = Array.isArray(type) ? type : [type];
            const queueName = (options === null || options === void 0 ? void 0 : options.queue) || (0, uuid_1.v4)();
            try {
                const { queue } = yield this.channel.assertQueue(queueName, {
                    exclusive: (_a = options === null || options === void 0 ? void 0 : options.exclusive) !== null && _a !== void 0 ? _a : !(options === null || options === void 0 ? void 0 : options.queue),
                    durable: true,
                    arguments: {
                        "x-dead-letter-exchange": this.options.connection.deadLetterExchange,
                        "x-message-ttl": 24 * 60 * 60 * 1000,
                        "x-max-length": this.QUEUE_MAX_LENGTH,
                        "x-overflow": "reject-publish",
                    },
                });
                for (const eventType of types) {
                    yield this.channel.bindQueue(queue, this.options.connection.exchange, eventType);
                    this.boundQueues.add(eventType);
                }
                yield this.channel.prefetch(this.options.consumer.prefetch);
                yield this.channel.consume(queue, (msg) => __awaiter(this, void 0, void 0, function* () {
                    var _a;
                    if (!msg)
                        return;
                    yield ((_a = this.messageHandler) === null || _a === void 0 ? void 0 : _a.handleMessage(msg, handler));
                }), { noAck: false });
                yield this.republishStoredMessages();
            }
            catch (error) {
                console.error(`Failed to subscribe to ${types.join(", ")}:`, error);
                throw error;
            }
        });
    }
    reconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            this.isInitialized = false;
            try {
                yield this.init();
            }
            catch (error) {
                console.error("Failed to reconnect:", error);
                setTimeout(() => this.reconnect(), this.RECONNECT_DELAY);
            }
        });
    }
    recreateChannel() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                if (!this.connection)
                    throw new Error("No connection available");
                this.channel = yield this.connection.createChannel();
                yield this.setupExchangesAndQueues();
                yield this.initializeHandlers();
                console.log("Channel recreated successfully");
            }
            catch (error) {
                console.error("Failed to recreate channel:", error);
                yield this.reconnect();
            }
        });
    }
    unsubscribe(consumerTag) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.channel) {
                throw new Error("Channel not initialized");
            }
            try {
                yield this.channel.cancel(consumerTag);
                console.log(`Unsubscribed consumer ${consumerTag}`);
            }
            catch (error) {
                console.error(`Failed to unsubscribe consumer ${consumerTag}:`, error);
                throw error;
            }
        });
    }
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.idempotencyStore.dispose();
                if (this.channel) {
                    yield this.channel.close();
                    this.channel = undefined;
                }
                if (this.connection) {
                    yield this.connection.close();
                    this.connection = undefined;
                }
                this.isInitialized = false;
                console.log("RabbitMQ EventBus closed successfully");
            }
            catch (error) {
                throw new errors_1.MessageProcessingError("Failed to close RabbitMQ connection", error);
            }
        });
    }
}
exports.RabbitMQEventBus = RabbitMQEventBus;
