"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const uuid_1 = require("uuid");
const src_1 = require("../src");
const errors_1 = require("../src/implementations/rabbitmq/errors");
function getEventHandler(mockObj, eventName) {
    var _a;
    const handler = (_a = mockObj.on.mock.calls.find((call) => call[0] === eventName)) === null || _a === void 0 ? void 0 : _a[1];
    if (!handler) {
        throw new Error(`Handler for event "${eventName}" not found.`);
    }
    return handler;
}
jest.mock('amqplib', () => ({
    connect: jest.fn(),
}));
describe('RabbitMQEventBus', () => {
    let eventBus;
    let mockConnection;
    let mockChannel;
    const defaultOptions = {
        connection: {
            url: 'amqp://localhost',
            exchange: 'test-exchange',
            exchangeType: 'topic',
            deadLetterExchange: 'test-dlx',
            retryExchange: 'test-retry'
        }
    };
    beforeEach(() => {
        // Setup mock channel
        mockChannel = {
            assertExchange: jest.fn().mockResolvedValue(undefined),
            assertQueue: jest.fn().mockResolvedValue({ queue: 'test-queue' }),
            bindQueue: jest.fn().mockResolvedValue(undefined),
            publish: jest.fn().mockResolvedValue(true),
            consume: jest.fn().mockImplementation((queue, callback) => {
                return Promise.resolve({ consumerTag: 'test-consumer' });
            }),
            prefetch: jest.fn().mockResolvedValue(undefined),
            ack: jest.fn().mockResolvedValue(undefined),
            nack: jest.fn().mockResolvedValue(undefined),
            cancel: jest.fn().mockResolvedValue(undefined),
            close: jest.fn().mockResolvedValue(undefined),
            on: jest.fn(),
        };
        // Setup mock connection
        mockConnection = {
            createChannel: jest.fn().mockResolvedValue(mockChannel),
            close: jest.fn().mockResolvedValue(undefined),
            on: jest.fn(),
        };
        require('amqplib').connect.mockResolvedValue(mockConnection);
        eventBus = new src_1.RabbitMQEventBus(defaultOptions);
    });
    afterEach(() => __awaiter(void 0, void 0, void 0, function* () {
        jest.clearAllMocks();
        if (eventBus) {
            yield eventBus.close();
        }
    }));
    describe('Initialization', () => {
        test('should initialize successfully', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
            expect(require('amqplib').connect).toHaveBeenCalledWith(defaultOptions.connection.url);
            expect(mockConnection.createChannel).toHaveBeenCalled();
            expect(mockChannel.assertExchange).toHaveBeenCalledWith(defaultOptions.connection.exchange, defaultOptions.connection.exchangeType, expect.any(Object));
        }));
        test('should handle initialization failure', () => __awaiter(void 0, void 0, void 0, function* () {
            eventBus.setDisableReconnection(true); // Désactive la reconnexion pour le test
            const error = new Error('Connection failed');
            require('amqplib').connect.mockRejectedValueOnce(error);
            yield expect(eventBus.init()).rejects.toThrow('Connection failed');
        }));
        test('should setup error handlers during initialization', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
            expect(mockConnection.on).toHaveBeenCalledWith('error', expect.any(Function));
            expect(mockConnection.on).toHaveBeenCalledWith('close', expect.any(Function));
            expect(mockChannel.on).toHaveBeenCalledWith('error', expect.any(Function));
            expect(mockChannel.on).toHaveBeenCalledWith('close', expect.any(Function));
        }));
    });
    describe('Publishing', () => {
        const testEvent = {
            type: 'test.event',
            data: { message: 'test' },
            metadata: {
                specversion: '1.0',
                id: (0, uuid_1.v4)(),
                time: new Date().toISOString(),
                source: 'test-source',
            },
        };
        beforeEach(() => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
        }));
        test('should publish message successfully', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.publish(testEvent);
            expect(mockChannel.publish).toHaveBeenCalledWith(defaultOptions.connection.exchange, testEvent.type, expect.any(Buffer), expect.objectContaining({
                messageId: expect.any(String),
                timestamp: expect.any(Number),
            }));
        }));
        test('should handle publish failure', () => __awaiter(void 0, void 0, void 0, function* () {
            mockChannel.publish.mockReturnValueOnce(false);
            yield expect(eventBus.publish(testEvent)).rejects.toThrow(errors_1.MessageProcessingError);
        }));
        test('should publish with custom options', () => __awaiter(void 0, void 0, void 0, function* () {
            const customOptions = {
                persistent: true,
                headers: { 'custom-header': 'value' }
            };
            yield eventBus.publish(testEvent, customOptions);
            expect(mockChannel.publish).toHaveBeenCalledWith(defaultOptions.connection.exchange, testEvent.type, expect.any(Buffer), expect.objectContaining({
                persistent: true,
                headers: expect.objectContaining({
                    'custom-header': 'value'
                })
            }));
        }));
    });
    describe('Subscribing', () => {
        const handler = jest.fn();
        beforeEach(() => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
        }));
        test('should subscribe to event successfully', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.subscribe('test.event', handler);
            expect(mockChannel.assertQueue).toHaveBeenCalled();
            expect(mockChannel.bindQueue).toHaveBeenCalled();
            expect(mockChannel.consume).toHaveBeenCalled();
        }));
        test('should handle multiple event types', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.subscribe(['test.event1', 'test.event2'], handler);
            const eventBindings = mockChannel.bindQueue.mock.calls.filter(call => call[2] === 'test.event1' || call[2] === 'test.event2');
            expect(eventBindings).toHaveLength(2);
        }));
        test('should handle message processing', () => __awaiter(void 0, void 0, void 0, function* () {
            const testMessage = {
                content: Buffer.from(JSON.stringify({ test: 'data' })),
                properties: {
                    messageId: 'test-id',
                    headers: {}
                },
                fields: {
                    routingKey: 'test.event',
                    exchange: 'test-exchange'
                }
            };
            // Simulate message consumption
            yield eventBus.subscribe('test.event', handler);
            const consumeCallback = mockChannel.consume.mock.calls[0][1];
            yield consumeCallback(testMessage);
            expect(handler).toHaveBeenCalled();
            expect(mockChannel.ack).toHaveBeenCalledWith(testMessage);
        }));
        test('should handle retryable errors', () => __awaiter(void 0, void 0, void 0, function* () {
            const handler = jest.fn().mockRejectedValueOnce(new errors_1.RetryableError('Retry me'));
            yield eventBus.subscribe('test.event', handler);
            const testMessage = {
                content: Buffer.from(JSON.stringify({ test: 'data' })),
                properties: {
                    messageId: 'test-id',
                    headers: {}
                },
                fields: {
                    routingKey: 'test.event',
                    exchange: 'test-exchange'
                }
            };
            const consumeCallback = mockChannel.consume.mock.calls[0][1];
            yield consumeCallback(testMessage);
            expect(mockChannel.publish).toHaveBeenCalledWith(defaultOptions.connection.retryExchange, 'test.event', expect.any(Buffer), expect.objectContaining({
                headers: expect.objectContaining({
                    'x-retry-count': 1
                })
            }));
        }));
    });
    describe('Dead Letter Handling', () => {
        test('should setup dead letter queue', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
            expect(mockChannel.assertExchange).toHaveBeenCalledWith(defaultOptions.connection.deadLetterExchange, 'topic', expect.any(Object));
            expect(mockChannel.assertQueue).toHaveBeenCalledWith(expect.stringContaining('dlq'), expect.any(Object));
        }));
        test('should handle non-retryable errors', () => __awaiter(void 0, void 0, void 0, function* () {
            const handler = jest.fn().mockRejectedValueOnce(new Error('Fatal error'));
            yield eventBus.init();
            yield eventBus.subscribe('test.event', handler);
            const testMessage = {
                content: Buffer.from(JSON.stringify({ test: 'data' })),
                properties: {
                    messageId: 'test-id',
                    headers: {}
                },
                fields: {
                    routingKey: 'test.event',
                    exchange: 'test-exchange'
                }
            };
            const consumeCallback = mockChannel.consume.mock.calls[0][1];
            yield consumeCallback(testMessage);
            expect(mockChannel.publish).toHaveBeenCalledWith(defaultOptions.connection.deadLetterExchange, 'dead.letter', expect.any(Buffer), expect.any(Object));
        }));
    });
    describe('Connection Management', () => {
        test('should handle reconnection on connection loss', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
            // Simulate connection loss
            const connectionError = new Error('Connection lost');
            const errorHandler = getEventHandler(mockConnection, 'error');
            yield errorHandler(connectionError);
            expect(require('amqplib').connect).toHaveBeenCalledTimes(2);
        }));
        test('should handle channel recreation', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
            // Simulate channel error
            const channelError = new Error('Channel error');
            const errorHandler = getEventHandler(mockChannel, 'error');
            yield errorHandler(channelError);
            expect(mockConnection.createChannel).toHaveBeenCalledTimes(2);
        }));
    });
    describe('Cleanup', () => {
        test('should close connection properly', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
            yield eventBus.close();
            expect(mockChannel.close).toHaveBeenCalled();
            expect(mockConnection.close).toHaveBeenCalled();
        }));
        test('should handle close errors', () => __awaiter(void 0, void 0, void 0, function* () {
            mockChannel.close.mockRejectedValueOnce(new Error('Close failed'));
            yield eventBus.init();
            yield expect(eventBus.close()).rejects.toThrow(errors_1.MessageProcessingError);
        }));
    });
    describe('Message Store', () => {
        test('should store unroutable messages', () => __awaiter(void 0, void 0, void 0, function* () {
            yield eventBus.init();
            // Simulate unroutable message
            const unroutableMessage = {
                content: Buffer.from('test'),
                properties: { messageId: 'test-id' },
                fields: {
                    exchange: 'test-exchange',
                    routingKey: 'unknown.route'
                }
            };
            const returnHandler = getEventHandler(mockChannel, 'return');
            returnHandler(unroutableMessage);
            // Vérifier que le message a été stocké (il faudrait exposer une méthode pour vérifier)
            // Cette vérification dépendra de votre implémentation spécifique
        }));
    });
    describe('Idempotency', () => {
        test('should handle duplicate messages', () => __awaiter(void 0, void 0, void 0, function* () {
            const handler = jest.fn();
            const messageId = 'duplicate-id';
            yield eventBus.init();
            yield eventBus.subscribe('test.event', handler);
            const testMessage = {
                content: Buffer.from(JSON.stringify({ test: 'data' })),
                properties: {
                    messageId,
                    headers: {}
                },
                fields: {
                    routingKey: 'test.event',
                    exchange: 'test-exchange'
                }
            };
            const consumeCallback = mockChannel.consume.mock.calls[0][1];
            // Premier traitement
            yield consumeCallback(testMessage);
            // Deuxième traitement du même message
            yield consumeCallback(testMessage);
            // Le handler ne devrait être appelé qu'une seule fois
            expect(handler).toHaveBeenCalledTimes(1);
        }));
    });
});
